@@ -10,7 +10,7 @@
 //! like `RefCell`.
 
 use std::borrow::Borrow;
-use std::cell::Cell;
+use std::cell::{Cell, RefCell};
 use std::collections::{HashMap, HashSet};
 use std::fmt::{Debug, Display};
 use std::hash::{Hash, Hasher};
@@ -23,40 +23,47 @@ use crate::thread_id::current_thread;
 struct Inner<T> {
     // id of thread from which the value can be accessed
     pinned_to: AtomicU64,
+    marking: Cell<u64>,
     val: T,
-    // the reference count
     strong_count: Cell<usize>,
 }
 
-static ID_NEXT: AtomicUsize = AtomicUsize::new(0);
+static NEXT_SENDRC_ID: AtomicUsize = AtomicUsize::new(0);
+static NEXT_PRE_SEND_ID: AtomicU64 = AtomicU64::new(1);
 
 /// Reference-counting pointer like `Rc<T>`, but which is `Send` if `T` is `Send`.
 ///
 /// After a `SendRc` is created, it is pinned to the current thread, and usable only in
-/// that thread. When sending a `SendRc` to different thread, you must first disable all
-/// the `SendRc`s that point to the same allocation, then send them, and finally reenable
-/// them in the new thread. They may again be sent to a different thread using the same
-/// process.
+/// that thread. Before sending a `SendRc` to different thread, you must call
+/// [`SendRc::pre_send()`] and use the returned `PreSend` to mark all the `SendRc`s that
+/// point to the same shared value for sending with a call to
+/// [`pre_send.mark_send()`](PreSend::mark_send) on each. This will make the values
+/// temporarily inaccessible.
+///
+/// When done with marking, you can obtain the `post_send` token with
+/// [`pre_send.ready()`](PreSend::ready), and restore access to values by calling
+/// [`post_send.sent()`](PostSend::sent). This process may be repeated sent the `SendRc`s
+/// to another thread later.
 ///
 /// ```
 /// # use std::cell::RefCell;
 /// # use sendable::SendRc;
-/// // create two SendRcs pointing to the same allocation
+/// // create two SendRcs pointing to the same shared value
 /// let mut r1 = SendRc::new(RefCell::new(1));
 /// let mut r2 = SendRc::clone(&r1);
 ///
 /// // prepare to ship them off to a different thread
-/// let mut pre_send = SendRc::pre_send();
-/// pre_send.disable(&mut r1); // r1 is unusable from this point
-/// pre_send.disable(&mut r2); // r2 is unusable from this point
-/// // ready() would panic on un-disabled SendRcs pointing to the allocation of r1/r2
+/// let pre_send = SendRc::pre_send();
+/// pre_send.mark_send(&mut r1); // r1 and r2 cannot be dereferenced from this point
+/// pre_send.mark_send(&mut r2);
+/// // ready() would panic if there were unmarked SendRcs pointing to the shared value
 /// let mut post_send = pre_send.ready();
 ///
 /// // move everything to a different thread
 /// std::thread::spawn(move || {
 ///     // both pointers are unusable here
-///     post_send.enable(&mut r1); // r1 is usable from this point
-///     post_send.enable(&mut r2); // r2 is usable from this point
+///     post_send.sent();
+///     // they are usable from this point
 ///     *r1.borrow_mut() += 1;
 ///     assert_eq!(*r2.borrow(), 2);
 /// })
@@ -64,41 +71,53 @@ static ID_NEXT: AtomicUsize = AtomicUsize::new(0);
 /// .unwrap();
 /// ```
 ///
-/// If the `SendRc`s are edges in a graph, you'll need to visit the whole graph before and
-/// after the migration to the new thread. In the pre-send phase you'll need to disable the
-/// pointer after visiting its neighbors, whereas in the post-send step you'll need to first
-/// re-enable the pointer and then visit the neighbors.
+/// Compared to `Rc`, tradeoffs of a `SendRc` are:
 ///
-/// Compared to `Rc`, tradeoffs are:
-///
-/// * `deref()`, `clone()`, and `drop()` requires a check that the pointer is not disabled,
-///    and a relaxed atomic load to check that we're accessing it from the correct thread.
-/// * takes up two machine words.
-/// * doesn't support weak pointers (though such support could be added).
+/// * `deref()`, `clone()`, and `drop()` require a check that the shared value is not
+///    marked for sending, and a check that we're accessing it from the correct thread.
+/// * a `SendRc` takes up two machine words.
+/// * it currently doesn't support weak pointers.
 pub struct SendRc<T> {
     ptr: NonNull<Inner<T>>,
     // Associate an non-changing id with each pointer so that we can track how many have
     // participated in migration. If malicious code forces wrap-around of the id, we're
-    // still sound because passing two SendRcs with the same id to `PreSend::disable()`
+    // still sound because passing two SendRcs with the same id to `PreSend::mark_send()`
     // will just cause `PreSend::ready()` to always fail and prevent migration.
     id: usize,
 }
 
 // Safety: SendRc can be sent between threads because we prohibit access to clone, drop,
 // and deref except from the thread they are pinned to. Access is granted only after all
-// pointers to the same allocation have been migrated to the new thread, which is why we
+// pointers to the same shared value have been migrated to the new thread, which is why we
 // can avoid requiring T: Sync.
 unsafe impl<T> Send for SendRc<T> where T: Send {}
 
+enum PinCheck {
+    Ok,
+    BadThread,
+    Marking,
+}
+
+impl PinCheck {
+    fn errmsg(&self) -> &'static str {
+        match self {
+            PinCheck::Ok => "no error",
+            PinCheck::BadThread => "SendRc accessed from wrong thread; call pre_send() first",
+            PinCheck::Marking => "access to SendRc that is about to be sent to a new thread",
+        }
+    }
+}
+
 impl<T> SendRc<T> {
     /// Constructs a new `SendRc<T>`.
     ///
     /// The newly created `SendRc` is only usable from the current thread. To send it to
-    /// another thread, you must call `pre_send()`, disable it, and re-enable it in the
+    /// another thread, you must call `pre_send()`, mark it, and re-enable it in the
     /// new thread.
     pub fn new(val: T) -> Self {
         let ptr = Box::into_raw(Box::new(Inner {
             pinned_to: AtomicU64::new(current_thread()),
+            marking: Cell::new(0),
             val,
             strong_count: Cell::new(1),
         }));
@@ -110,12 +129,12 @@ impl<T> SendRc<T> {
             // unwrap: we have a valid box, its pointer is not null (rustc eliminates the
             // check, https://godbolt.org/z/dsYPxxMWo)
             ptr: NonNull::new(ptr).unwrap(),
-            id: ID_NEXT.fetch_add(1, Ordering::Relaxed),
+            id: NEXT_SENDRC_ID.fetch_add(1, Ordering::Relaxed),
         }
     }
 
     fn inner(&self) -> &Inner<T> {
-        // Safety: Allocation is valid at least as long as self
+        // Safety: Shared Value is valid at least as long as self
         unsafe { self.ptr.as_ref() }
     }
 
@@ -125,75 +144,78 @@ impl<T> SendRc<T> {
     }
 
     #[inline]
-    fn check_pinned_and_valid(&self) -> Result<(), &'static str> {
-        if self.ptr == NonNull::dangling() {
-            return Err("attempt to use disabled SendRc");
+    fn check_pinned(&self) -> PinCheck {
+        let inner = self.inner();
+        if inner.pinned_to.load(Ordering::Relaxed) != current_thread() {
+            return PinCheck::BadThread;
         }
-        if self.inner().pinned_to.load(Ordering::Relaxed) != current_thread() {
-            return Err("attempt to use SendRc from different thread; call pre_send() first");
+        if inner.marking.get() != 0 {
+            return PinCheck::Marking;
         }
-        Ok(())
+        PinCheck::Ok
     }
 
-    fn assert_pinned_and_valid(&self, op: &str) {
-        if let Err(msg) = self.check_pinned_and_valid() {
-            panic!("{op}: {msg}");
+    fn assert_pinned(&self, op: &str) {
+        match self.check_pinned() {
+            PinCheck::Ok => {}
+            check @ (PinCheck::BadThread | PinCheck::Marking) => {
+                panic!("{op}: {}", check.errmsg());
+            }
         }
     }
 
     /// Prepare to send `SendRc`s of this type to a different thread.
     ///
-    /// Before moving a `SendRc` to a different thread, you must disable it as well as all
-    /// other `SendRc`s pointing to the same allocation:
+    /// Before moving a `SendRc` to a different thread, you must call
+    /// [`mark_send()`](PreSend::mark_send) on that pointer, as well as on all other
+    /// `SendRc`s pointing to the same shared value.
     ///
     /// ```
     /// # use std::cell::RefCell;
     /// # use sendable::SendRc;
     /// let mut r1 = SendRc::new(RefCell::new(1));
     /// let mut r2 = SendRc::clone(&r1);
-    /// let mut pre_send = SendRc::pre_send();
-    /// pre_send.disable(&mut r1);
-    /// pre_send.disable(&mut r2);
+    /// let pre_send = SendRc::pre_send();
+    /// pre_send.mark_send(&mut r1);
+    /// pre_send.mark_send(&mut r2);
     /// let mut post_send = pre_send.ready();
-    /// // send, r1, and r2 can now be send to a different thread, and re-enabled
-    /// // by calling post_send.enable(&mut r1) and post_send.enable(&mut r2)
-    /// # post_send.enable(&mut r1); post_send.enable(&mut r2); // avoid panic in doctest
+    /// // post_send, r1, and r2 can now be send to a different thread, and re-enabled
+    /// // by calling post_send.sent()
+    /// # post_send.sent(); // avoid panic in doctest
     /// ```
     pub fn pre_send() -> PreSend<T> {
         PreSend {
-            disabled: HashMap::new(),
+            marked: Default::default(),
+            pre_send_id: NEXT_PRE_SEND_ID.fetch_add(1, Ordering::Relaxed),
         }
     }
 
     /// Prepare to send a fixed collection of `SendRc`s to a different thread.
     ///
-    /// Calls `SendRc::pre_send()`, then `disable()` on the provided `SendRc`s, then
-    /// finishes the pre-send phase with a call to `ready()`. Returns the `PostSend` token
-    /// which to send to the new thread along with the `SendRc`s and use to re-enable the
-    /// `SendRc`s.
+    /// Calls `SendRc::pre_send()`, followed by `mark_send()` on the provided `SendRc`s,
+    /// and finally returns the post-send token returned by the call to
+    /// `pre_send.ready()`. Useful for one-shot move of `SendRc`s which are easily
+    /// traversable.
     ///
-    /// Panics if there are allocations pointed to by `SendRc`s in `all` which have extra
-    /// `SendRc`s not included in `all`.
+    /// Panics if there are shared values pointed to by `SendRc`s in `all` which have
+    /// extra `SendRc`s not included in `all`.
     pub fn pre_send_ready<'a>(all: impl IntoIterator<Item = &'a mut Self>) -> PostSend<T>
     where
         T: 'a,
     {
-        let mut send = Self::pre_send();
-        send.disable_many(all);
-        send.ready()
+        let pre_send = Self::pre_send();
+        for send_rc in all {
+            pre_send.mark_send(send_rc);
+        }
+        pre_send.ready()
     }
 
-    /// Returns true if this `SendRc` has been disabled for sending to a new thread.
-    pub fn is_sendrc_disabled(&self) -> bool {
-        self.ptr == NonNull::dangling()
-    }
-
-    /// Returns the number of pointers to this allocation.
+    /// Returns the number of `SendRc`s pointing to this shared value.
     ///
     /// Panics when invoked from a different thread than the one the `SendRc` was created
     /// in or last migrated to.
     pub fn strong_count(this: &Self) -> usize {
-        this.assert_pinned_and_valid("SendRc::strong_count()");
+        this.assert_pinned("SendRc::strong_count()");
         this.inner().strong_count.get()
     }
 
@@ -202,7 +224,7 @@ impl<T> SendRc<T> {
     /// Panics when invoked from a different thread than the one the `SendRc` was created
     /// in or last migrated to.
     pub fn try_unwrap(this: Self) -> Result<T, Self> {
-        this.assert_pinned_and_valid("SendRc::try_unwrap()");
+        this.assert_pinned("SendRc::try_unwrap()");
         if this.inner().strong_count.get() == 1 {
             // Safety: refcount is 1, so it's just us, and the pointer was obtained using
             // Box::into_raw().
@@ -213,13 +235,13 @@ impl<T> SendRc<T> {
         }
     }
 
-    /// Returns a mutable reference into the given `SendRc`, if there are no other `SendRc`
-    /// pointers to the same allocation.
+    /// Returns a mutable reference into the given `SendRc`, if there no other `SendRc`s
+    /// point to the same shared value.
     ///
     /// Panics when invoked from a different thread than the one the `SendRc` was created
     /// in or last migrated to.
     pub fn get_mut(this: &mut Self) -> Option<&mut T> {
-        this.assert_pinned_and_valid("SendRc::get_mut()");
+        this.assert_pinned("SendRc::get_mut()");
         if this.inner().strong_count.get() == 1 {
             // Safety: we've checked that refcount is 1
             unsafe { Some(&mut this.inner_mut().val) }
@@ -228,56 +250,84 @@ impl<T> SendRc<T> {
         }
     }
 
-    /// Returns true if the two `SendRc`s point to the same allocation.
+    /// Returns true if this and the other `SendRc`s point to the same shared value.
     pub fn ptr_eq(this: &Self, other: &Self) -> bool {
         this.ptr == other.ptr
     }
 }
 
-/// Handle for disabling the `SendRc`s before they are sent to another thread.
+/// Handle for marking the `SendRc`s before they are sent to another thread.
 ///
-/// This handle cannot be sent to a different thread; when done with calls to `disable()`,
-/// invoke `ready()` to obtain a token that can be moved to the other thread to re-enable
-/// the `SendRc`s.
+/// This handle is not `Send`; when done with calls to `mark_send()`, invoke `ready()` to
+/// obtain a token that can be moved to the other thread to re-enable the `SendRc`s.
 pub struct PreSend<T> {
-    disabled: HashMap<usize, NonNull<Inner<T>>>,
+    marked: RefCell<HashMap<usize, NonNull<Inner<T>>>>,
+    pre_send_id: u64,
 }
 
 impl<T> PreSend<T> {
-    /// Make `send_rc` temporarily unusable so it can be sent to another thread.
+    /// Make the value pointed to by `send_rc` temporarily unreachable, so it can be
+    /// safely sent to another thread.
     ///
-    /// After this call it is no longer possible to deref, clone, or drop this `SendRc`.
+    /// After this call it is no longer possible to deref, clone, or drop either this
+    /// `SendRc` or any other that points to the same shared value. However, the shared
+    /// value can be accessed reachable through the return value of this method, which may
+    /// be called again on the same `SendRc`. This allows reaching other `SendRc`s while
+    /// traversing the graph.
     ///
-    /// After disabling a `SendRc`, you must disable other `SendRc`s pointing to the same
-    /// allocation. This needs to be done before the call to `ready()`.
+    /// It is allowed to call this method with `SendRc`s pointing to different shared
+    /// values. After marking a so far unmarked shared value through a `SendRc`, prior to
+    /// invoking [`ready()`](PreSend::ready) you must make sure to also mark all the other
+    /// `SendRc`s pointing to that shared value.
+    ///
+    /// Panics when invoked from a different thread than the one the `SendRc` was created
+    /// in or last migrated to. Also panics when passed a `send_rc` that was already
+    /// marked by a different `PreSend` handle.
+    pub fn mark_send<'a>(&'a self, send_rc: &'a mut SendRc<T>) -> &'a T {
+        match send_rc.check_pinned() {
+            PinCheck::Ok => {
+                send_rc.inner().marking.set(self.pre_send_id);
+            }
+            check @ PinCheck::BadThread => panic!("PreSend::mark_send(): {}", check.errmsg()),
+            PinCheck::Marking => {
+                // Don't allow mark_send() from a rogue PreSend because we return a
+                // reference tied to the lifetime of self, and depend on self being
+                // consumed before it's sent to the new thread.
+                if send_rc.inner().marking.get() != self.pre_send_id {
+                    panic!("PreSend::mark_send(): call from different PreSend");
+                }
+            }
+        }
+        self.marked.borrow_mut().insert(send_rc.id, send_rc.ptr);
+        &send_rc.inner().val
+    }
+
+    /// Returns true if `send_rc` was already marked for sending by this `PreSend` handle.
     ///
     /// Panics when invoked from a different thread than the one the `SendRc` was created
     /// in or last migrated to.
-    pub fn disable(&mut self, send_rc: &mut SendRc<T>) {
-        if self.disabled.contains_key(&send_rc.id) {
-            // make calling disable() twice a no-op
-            return;
-        }
-        // If send_rc.ptr is dangling, it means the send_rc was already disabled in
-        // another SendRc.  This is a programming error, and will be caught by
-        // assert_pinned_and_valid().
-        send_rc.assert_pinned_and_valid("PreSend::disable()");
-        self.disabled.insert(send_rc.id, send_rc.ptr);
-        send_rc.ptr = NonNull::dangling();
-    }
-
-    /// Make multiple `SendRc`s temporarily unusable as if with `disable()`.
-    pub fn disable_many<'a>(&mut self, many: impl IntoIterator<Item = &'a mut SendRc<T>>)
-    where
-        T: 'a,
-    {
-        for send_rc in many {
-            self.disable(send_rc);
+    pub fn is_marked(&self, send_rc: &SendRc<T>) -> bool {
+        match send_rc.check_pinned() {
+            PinCheck::Ok => false,
+            PinCheck::Marking => self.marked.borrow().contains_key(&send_rc.id),
+            check @ PinCheck::BadThread => panic!("PreSend::is_marked(): {}", check.errmsg()),
         }
     }
 
-    /// Returns true if there are no allocations whose `SendRc`s we've disabled, but which
-    /// have outstanding `SendRc`s we haven't yet disabled.
+    /// Returns true if the shared value this `send_rc` points to was marked for sending.
+    ///
+    /// Panics when invoked from a different thread than the one the `SendRc` was created
+    /// in or last migrated to.
+    pub fn is_shared_value_marked(&self, send_rc: &SendRc<T>) -> bool {
+        match send_rc.check_pinned() {
+            PinCheck::Ok => false,
+            PinCheck::Marking => true,
+            check @ PinCheck::BadThread => panic!("PreSend::is_marked(): {}", check.errmsg()),
+        }
+    }
+
+    /// Returns true if there are no unmarked `SendRc`s pointing to the shared values
+    /// whose `SendRc`s we've marked.
     ///
     /// If this returns true, it means [`ready()`](PreSend::ready) will succeed without
     /// panic.
@@ -290,137 +340,85 @@ impl<T> PreSend<T> {
     /// let mut r2 = SendRc::clone(&r1);
     /// let mut q1 = SendRc::new(RefCell::new(1));
     /// let mut q2 = SendRc::clone(&q1);
-    /// let mut pre_send = SendRc::pre_send();
-    /// pre_send.disable(&mut r1);
-    /// assert!(pre_send.all_disabled() == false); // r2 not disabled
-    /// pre_send.disable(&mut r2);
-    /// assert!(pre_send.all_disabled() == true); // r1/r2 allocation fully disabled, q1/q2 doesn't participate
-    /// pre_send.disable(&mut q1);
-    /// assert!(pre_send.all_disabled() == false); // r1/r2 ok, but q2 not disabled
-    /// pre_send.disable(&mut q2);
-    /// assert!(pre_send.all_disabled() == true); // both r1/r2 allocations and q1/q2 allocations fully disabled
+    /// let pre_send = SendRc::pre_send();
+    /// pre_send.mark_send(&mut r1);
+    /// assert!(pre_send.all_marked() == false); // r2 still unmarked
+    /// pre_send.mark_send(&mut r2);
+    /// assert!(pre_send.all_marked() == true); // r1/r2 shared value fully marked, q1/q2 not involved
+    /// pre_send.mark_send(&mut q1);
+    /// assert!(pre_send.all_marked() == false); // r1/r2 ok, but q2 unmarked
+    /// pre_send.mark_send(&mut q2);
+    /// assert!(pre_send.all_marked() == true); // both r1/r2 and q1/q2 shared values fully marked
     /// # std::mem::forget([r1, r2, q1, q2]);
     /// ```
-    pub fn all_disabled(&self) -> bool {
-        // Count how many `SendRc`s point to each allocation.
+    pub fn all_marked(&self) -> bool {
+        // Count how many `SendRc`s point to each shared value.
         let ptr_sendrc_cnt: HashMap<_, usize> =
-            self.disabled
+            self.marked
+                .borrow()
                 .values()
                 .fold(HashMap::new(), |mut map, &ptr| {
                     *map.entry(ptr).or_default() += 1;
                     map
                 });
         ptr_sendrc_cnt.into_iter().all(|(ptr, cnt)| {
-            // Safety: allocation is valid because there is at least 1 SendRc pointing to
+            // Safety: shared value is valid because there is at least 1 SendRc pointing to
             // it. We may access it from this thread because PreSend isn't Send.
             let inner = unsafe { &*ptr.as_ptr() };
             cnt == inner.strong_count.get()
         })
     }
 
-    /// Returns a [`PostSend`] token that can proves all `SendRc`s have been disabled.
+    /// Assert that all `SendRc`s pointing to shared values that are to be moved to a new
+    /// thread have been marked and return a [`PostSend`] token as proof of this check.
     ///
-    /// This token can be sent to another thread and used to re-enable the `SendRc`s
-    /// there.
+    /// The returned token can be sent to another thread along with the `SendRc`s, and
+    /// used to re-enable them with a call to [`sent()`](PostSend::sent).
     ///
-    /// Panics if there are outstanding `SendRc`s we haven't yet disabled, i.e. if
-    /// `all_disabled()` would return false.
+    /// Panics if the above assertion is untrue, i.e. if
+    /// [`all_marked()`](PreSend::all_marked) would return false.
     pub fn ready(self) -> PostSend<T> {
-        if !self.all_disabled() {
-            panic!("PreSend::ready() called before all SendRcs have been disabled");
+        if !self.all_marked() {
+            panic!("PreSend::ready() called before all SendRcs have been marked");
         }
-        // Pin allocations to a non-existent thread id, so that enable() can detect the
+        let ptrs: HashSet<_> = self.marked.into_inner().into_values().collect();
+        // Pin shared values to a non-existent thread id, so that enable() can detect the
         // new thread from which we are called (even if that new thread is the current
         // thread again).
-        for &ptr in self.disabled.values() {
-            // Safety: ptr belongs to a SendRc, so it's valid
+        for &ptr in &ptrs {
+            // Safety: ptr belongs to at least one SendRc, so it's valid
             let inner = unsafe { &*ptr.as_ptr() };
             inner.pinned_to.store(0, Ordering::Relaxed);
         }
-        PostSend {
-            disabled: self.disabled,
-            enabled: HashSet::new(),
-            new_thread: 0,
-        }
+        PostSend { ptrs }
     }
 }
 
 /// Handle for enabling the `SendRc`s after they are sent to another thread.
 ///
 /// Since `PostSend` can only be obtained via [`PreSend::ready()`], possessing a
-/// `PostSend` serves as proof that all `SendRc`s belonging to the allocations involved
-/// in the move have been disabled.
+/// `PostSend` serves as proof that all `SendRc`s pointing to the shared values involved in
+/// the move have been marked.
 #[must_use]
 pub struct PostSend<T> {
-    disabled: HashMap<usize, NonNull<Inner<T>>>,
-    enabled: HashSet<usize>,
-    new_thread: u64,
+    ptrs: HashSet<NonNull<Inner<T>>>,
 }
 
-// Safety: pointers to allocations can be sent to a new thread because PreSend::ready()
-// checked that all their SendRcs have been disabled.
+// Safety: pointers to shared values can be sent to a new thread because PreSend::ready()
+// checked that all their SendRcs have been marked.
 unsafe impl<T> Send for PostSend<T> where T: Send {}
 
 impl<T> PostSend<T> {
-    /// Make `send_rc` usable again after having moved it to a new thread.
-    pub fn enable(&mut self, send_rc: &mut SendRc<T>) {
+    /// Re-enable all pointers involved in the move and make their data accessible from
+    /// this thread.
+    pub fn sent(self) {
         let current_thread = current_thread();
-        // Disallow calling enable() from different threads in sequence, even if that
-        // would be sound. If the user needs multiple threads, they can create multiple
-        // PreSend/PostSend.
-        match self.new_thread {
-            0 => self.new_thread = current_thread,
-            id if id == current_thread => {}
-            _ => panic!("PostSend::enable() called from more than one thread"),
-        }
-        if self.enabled.contains(&send_rc.id) {
-            // make calling enable() twice a no-op
-            return;
-        }
-        if send_rc.ptr != NonNull::dangling() {
-            panic!("PostSend::enable() called on a non-disabled SendRc");
-        }
-        // This will panic if the SendRc was disabled by another PreSend.
-        let ptr = self
-            .disabled
-            .get(&send_rc.id)
-            .copied()
-            .expect("PostSend::enable() called on a SendRc disabled elsewhere");
-        // Safety: SendRc guarantees that the pointer points to valid allocation.
-        let inner = unsafe { &*ptr.as_ptr() };
-        // Check previously pinned thread in case someone sends PostSend to another
-        // thread, enables a pointer, and then sends it to a third thread and enables
-        // another one.
-        let old_pinned_to = inner.pinned_to.load(Ordering::Relaxed);
-        if old_pinned_to == 0 {
-            // We can get away with load+store without CAS because this can't happen in
-            // parallel, since we take &mut self.
+        for ptr in self.ptrs {
+            // Safety: ptr belongs to at least one SendRc, so it's valid
+            let inner = unsafe { &*ptr.as_ptr() };
             inner.pinned_to.store(current_thread, Ordering::Relaxed);
-        } else {
-            // Since we check that enable() is always called from the same thread, it
-            // shouldn't be possible for the allocation to be pinned to some other thread.
-            assert_eq!(old_pinned_to, current_thread);
+            inner.marking.set(0);
         }
-        send_rc.ptr = ptr;
-        self.enabled.insert(send_rc.id);
-    }
-
-    /// Make multiple `SendRc`s usable again as if with `enable()`.
-    pub fn enable_many<'a>(&mut self, many: impl IntoIterator<Item = &'a mut SendRc<T>>)
-    where
-        T: 'a,
-    {
-        for send_rc in many {
-            self.enable(send_rc);
-        }
-    }
-
-    /// Returns true if all the `SendRc`s have been enabled.
-    ///
-    /// Useful for asserting that the post-send traversal hasn't missed a `SendRc`
-    /// somewhere.
-    pub fn all_enabled(&self) -> bool {
-        self.enabled.len() == self.disabled.len()
     }
 }
 
@@ -440,14 +438,14 @@ impl<T> Deref for SendRc<T> {
     type Target = T;
 
     fn deref(&self) -> &T {
-        self.assert_pinned_and_valid("SendRc::deref()");
+        self.assert_pinned("SendRc::deref()");
         &self.inner().val
     }
 }
 
 impl<T> Clone for SendRc<T> {
     fn clone(&self) -> Self {
-        self.assert_pinned_and_valid("SendRc::clone()");
+        self.assert_pinned("SendRc::clone()");
         self.inner()
             .strong_count
             .set(self.inner().strong_count.get() + 1);
@@ -460,8 +458,8 @@ impl<T> Drop for SendRc<T> {
         // Instead of panicking immediately, check whether we're in the correct thread and
         // leak the value if we're not. Then panic, but only if we're not already
         // panicking, because panic-inside-panic aborts the program and breaks unit tests.
-        match self.check_pinned_and_valid() {
-            Ok(()) => {
+        match self.check_pinned() {
+            PinCheck::Ok => {
                 let refcnt = self.inner().strong_count.get();
                 if refcnt == 1 {
                     unsafe {
@@ -471,9 +469,9 @@ impl<T> Drop for SendRc<T> {
                     self.inner().strong_count.set(refcnt - 1);
                 }
             }
-            Err(msg) => {
+            check @ (PinCheck::BadThread | PinCheck::Marking) => {
                 if !std::thread::panicking() {
-                    panic!("SendRc::drop(): {msg}");
+                    panic!("SendRc::drop(): {}", check.errmsg());
                 }
             }
         }
@@ -524,62 +522,6 @@ impl<T: Hash> Hash for SendRc<T> {
     }
 }
 
-/// Common trait for `PreSend` and `PostSend`, allowing common code for traversal of
-/// `SendRc`s to disable/enable them.
-///
-/// When `SendRc`s form a graph, you'll need to visit them before and after the `Send`,
-/// and this trait is meant to help with making the code generic:
-///
-/// ```no_run
-/// # use sendable::{SendRc, send_rc::SendAction};
-/// # use std::cell::RefCell;
-/// struct Node {
-///     neighbor: SendRc<RefCell<Node>>,
-///     // data: ...
-/// }
-///
-/// impl Node {
-///     fn migrate(&mut self, action: &mut impl SendAction<RefCell<Node>>) {
-///         if action.will_disable() {
-///             self.neighbor.borrow_mut().migrate(action);
-///             action.apply(&mut self.neighbor);
-///         } else {
-///             action.apply(&mut self.neighbor);
-///             self.neighbor.borrow_mut().migrate(action);
-///         }
-///     }
-/// }
-/// ```
-pub trait SendAction<T> {
-    /// Returns true if the action will disable the pointer, i.e. if the action is
-    /// `PreSend`.
-    fn will_disable(&self) -> bool;
-
-    /// Calls `PreSend::disable()` or `PostSend::enable()` depending on the
-    /// implementation.
-    fn apply(&mut self, send_rc: &mut SendRc<T>);
-}
-
-impl<T> SendAction<T> for PreSend<T> {
-    fn will_disable(&self) -> bool {
-        true
-    }
-
-    fn apply(&mut self, send_rc: &mut SendRc<T>) {
-        self.disable(send_rc);
-    }
-}
-
-impl<T> SendAction<T> for PostSend<T> {
-    fn will_disable(&self) -> bool {
-        false
-    }
-
-    fn apply(&mut self, send_rc: &mut SendRc<T>) {
-        self.enable(send_rc);
-    }
-}
-
 #[cfg(test)]
 mod tests {
     use std::cell::RefCell;
@@ -639,10 +581,10 @@ mod tests {
     fn ok_send() {
         let mut r1 = SendRc::new(RefCell::new(1));
         let mut r2 = SendRc::clone(&r1);
-        let mut post_send = SendRc::pre_send_ready([&mut r1, &mut r2]);
+        let post_send = SendRc::pre_send_ready([&mut r1, &mut r2]);
 
         std::thread::spawn(move || {
-            post_send.enable_many([&mut r1, &mut r2]);
+            post_send.sent();
             *r1.borrow_mut() += 1;
             assert_eq!(*r2.borrow(), 2);
         })
@@ -650,26 +592,6 @@ mod tests {
         .unwrap();
     }
 
-    #[test]
-    fn send_and_return() {
-        let mut r1 = SendRc::new(RefCell::new(1));
-        let mut r2 = SendRc::clone(&r1);
-        let mut post_send = SendRc::pre_send_ready([&mut r1, &mut r2]);
-        let (mut post_send, mut r1, mut r2) = std::thread::spawn(move || {
-            post_send.enable_many([&mut r1, &mut r2]);
-            assert!(post_send.all_enabled());
-            *r1.borrow_mut() += 1;
-            assert_eq!(*r2.borrow(), 2);
-            let post_send = SendRc::pre_send_ready([&mut r1, &mut r2]);
-            (post_send, r1, r2)
-        })
-        .join()
-        .unwrap();
-        post_send.enable_many([&mut r1, &mut r2]);
-        assert!(post_send.all_enabled());
-        assert_eq!(*r2.borrow(), 2);
-    }
-
     #[test]
     #[should_panic]
     fn missing_pre_send_drop() {
@@ -698,22 +620,22 @@ mod tests {
     fn incomplete_pre_send() {
         let mut r1 = SendRc::new(RefCell::new(1));
         let _r2 = SendRc::clone(&r1);
-        let mut pre_send = SendRc::pre_send();
-        pre_send.disable(&mut r1);
-        let _ = pre_send.ready(); // panics because we didn't disable _r2
+        let pre_send = SendRc::pre_send();
+        pre_send.mark_send(&mut r1);
+        let _ = pre_send.ready(); // panics because we didn't mark _r2
     }
 
     #[test]
-    #[should_panic = "before all SendRcs have been disabled"]
-    fn incomplete_pre_send_other_allocation() {
+    #[should_panic = "before all SendRcs have been marked"]
+    fn incomplete_pre_send_other_shared_value() {
         let mut r1 = SendRc::new(RefCell::new(1));
         let mut r2 = SendRc::clone(&r1);
         let mut q1 = SendRc::new(RefCell::new(1));
         let _q2 = SendRc::clone(&q1);
-        let mut pre_send = SendRc::pre_send();
-        pre_send.disable(&mut r1);
-        pre_send.disable(&mut r2);
-        pre_send.disable(&mut q1);
+        let pre_send = SendRc::pre_send();
+        pre_send.mark_send(&mut r1);
+        pre_send.mark_send(&mut r2);
+        pre_send.mark_send(&mut q1);
         let _ = pre_send.ready(); // _q2 is missing
     }
 
@@ -722,129 +644,54 @@ mod tests {
     fn faked_pre_send_count_reusing_same_ptr() {
         let mut r1 = SendRc::new(RefCell::new(1));
         let _r2 = SendRc::clone(&r1);
-        let mut pre_send = SendRc::pre_send();
-        // disabling the same SendRc twice won't fool us into thinking all instances were
-        // disabled
-        pre_send.disable(&mut r1);
-        pre_send.disable(&mut r1);
+        let pre_send = SendRc::pre_send();
+        // marking the same SendRc twice won't fool us into thinking all SendRcs were
+        // marked
+        pre_send.mark_send(&mut r1);
+        pre_send.mark_send(&mut r1);
         let _ = pre_send.ready();
     }
 
     #[test]
-    #[should_panic = "attempt to use disabled"]
-    fn disable_same_sendrc_in_different_presend() {
+    #[should_panic = "call from different PreSend"]
+    fn mark_same_sendrc_in_different_presend() {
         let mut r = SendRc::new(RefCell::new(1));
-        let mut pre_send1 = SendRc::pre_send();
-        pre_send1.disable(&mut r);
-        let mut pre_send2 = SendRc::pre_send();
-        pre_send2.disable(&mut r);
+        let pre_send1 = SendRc::pre_send();
+        pre_send1.mark_send(&mut r);
+        let pre_send2 = SendRc::pre_send();
+        pre_send2.mark_send(&mut r);
         std::mem::forget(r); // prevent panic on drop
     }
 
     #[test]
-    fn enable_same_send_diff_threads() {
-        let mut r1 = SendRc::new(RefCell::new(1));
-        let mut r2 = SendRc::clone(&r1);
-        let mut post_send = SendRc::pre_send_ready([&mut r1, &mut r2]);
-        let mut post_send = std::thread::spawn(move || {
-            post_send.enable(&mut r1);
-            post_send
-        })
-        .join()
-        .unwrap();
-        let result = std::thread::spawn(move || {
-            post_send.enable(&mut r2);
-        })
-        .join();
-        assert!(result.is_err());
-    }
-
-    #[test]
-    fn enable_half_way() {
-        let state = Arc::new(Mutex::new(0));
-        let mut r1 = SendRc::new(RefCell::new(1));
-        let mut r2 = SendRc::clone(&r1);
-        let mut post_send = SendRc::pre_send_ready([&mut r1, &mut r2]);
-        let result = std::thread::spawn({
-            let state = state.clone();
-            move || {
-                *state.lock().unwrap() = 1;
-                post_send.enable(&mut r1);
-                *state.lock().unwrap() = 2;
-                *r1.borrow_mut() = 2;
-                *state.lock().unwrap() = 3;
-                let _ = &*r2; // should panic
-                *state.lock().unwrap() = 4;
-            }
-        })
-        .join();
-        assert!(result.is_err());
-        assert_eq!(*state.lock().unwrap(), 3);
-    }
-
-    #[test]
-    fn enable_undisabled() {
-        let state = Arc::new(Mutex::new(0));
-        let mut r1 = SendRc::new(RefCell::new(1));
-        let mut post_send = SendRc::pre_send_ready([&mut r1]);
-        let result = std::thread::spawn({
-            let state = state.clone();
-            move || {
-                *state.lock().unwrap() = 1;
-                post_send.enable(&mut r1);
-                let mut rogue = SendRc::clone(&r1);
-                *state.lock().unwrap() = 2;
-                post_send.enable(&mut rogue); // should panic
-                *state.lock().unwrap() = 3;
-            }
-        })
-        .join();
-        assert!(result.is_err());
-        assert_eq!(*state.lock().unwrap(), 2);
-    }
-
-    #[test]
-    fn enable_twice() {
-        let mut r1 = SendRc::new(RefCell::new(1));
-        let mut post_send = SendRc::pre_send_ready([&mut r1]);
-        std::thread::spawn(move || {
-            post_send.enable(&mut r1);
-            post_send.enable(&mut r1);
-            post_send.enable(&mut r1);
-        })
-        .join()
-        .unwrap();
-    }
-
-    #[test]
-    fn disable_twice_good() {
+    fn mark_twice_good() {
         let mut r1 = SendRc::new(RefCell::new(1));
         let mut r2 = SendRc::new(RefCell::new(1));
-        let mut pre_send = SendRc::pre_send();
-        pre_send.disable(&mut r1);
-        pre_send.disable(&mut r1);
-        pre_send.disable(&mut r2);
+        let pre_send = SendRc::pre_send();
+        pre_send.mark_send(&mut r1);
+        pre_send.mark_send(&mut r1);
+        pre_send.mark_send(&mut r2);
         let _ = pre_send.ready();
         // avoid panic on drop
         std::mem::forget([r1, r2]);
     }
 
     #[test]
-    fn disable_twice_bad() {
+    fn mark_twice_bad() {
         let state = Arc::new(Mutex::new(0));
         let result = std::thread::spawn({
             let state = state.clone();
             move || {
                 let mut r1 = SendRc::new(RefCell::new(1));
                 let mut r2 = SendRc::new(RefCell::new(1));
-                let mut pre_send1 = SendRc::pre_send();
-                let mut pre_send2 = SendRc::pre_send();
+                let pre_send1 = SendRc::pre_send();
+                let pre_send2 = SendRc::pre_send();
                 *state.lock().unwrap() = 1;
-                pre_send1.disable(&mut r1);
+                pre_send1.mark_send(&mut r1);
                 *state.lock().unwrap() = 2;
-                pre_send1.disable(&mut r2);
+                pre_send1.mark_send(&mut r2);
                 *state.lock().unwrap() = 3;
-                pre_send2.disable(&mut r2); // panic
+                pre_send2.mark_send(&mut r2); // panic
                 *state.lock().unwrap() = 4;
             }
         })
@@ -852,19 +699,4 @@ mod tests {
         assert!(result.is_err());
         assert_eq!(*state.lock().unwrap(), 3);
     }
-
-    #[test]
-    #[should_panic = "enable() called from more than one thread"]
-    fn send_from_diff_threads() {
-        let mut a = SendRc::new(RefCell::new(1));
-        let mut b = SendRc::clone(&a);
-        let mut post_send = SendRc::pre_send_ready([&mut a, &mut b]);
-        let mut post_send = std::thread::spawn(move || {
-            post_send.enable(&mut a);
-            post_send
-        })
-        .join()
-        .unwrap();
-        post_send.enable(&mut b);
-    }
 }
